from samson.utilities.bytes import Bytes

class RC4(object):
    """
    Rivest Cipher 4 (RC4)

    Broken stream ciphers with large, initial-keystream biases.
    """

    def __init__(self, key: bytes):
        """
        Parameters:
            key (bytes): Key (40-2040 bits).
        """
        self.key = key
        self.S = self.key_schedule(key)
        self.i = 0
        self.j = 0


    def __repr__(self):
        return f"<RC4: key={self.key}, S={self.S}, i={self.i}, j={self.j}>"

    def __str__(self):
        return self.__repr__()



    def key_schedule(self, key: bytes) -> list:
        """
        Prepares the internal state using the key.

        Parameters:
            key (bytes): Key.
        
        Returns:
            list: State parameter `S`.
        """
        key_length = len(key)
        S = []
        for i in range(256):
            S.append(i)

        j = 0
        for i in range(256):
            j = (j + S[i] + key[i % key_length]) % 256
            S[i], S[j] = S[j], S[i]
        
        return S


    def generate(self, length: int) -> Bytes:
        """
        Generates `length` of keystream.

        Parameters:
            length (int): Desired length of keystream in bytes.
        
        Returns:
            Bytes: Keystream.
        """
        keystream = Bytes(b'')
        
        for _ in range(length):
            self.i = (self.i + 1) % 256
            self.j = (self.j + self.S[self.i]) % 256
            self.S[self.i], self.S[self.j] = self.S[self.j], self.S[self.i]
            keystream += bytes([self.S[(self.S[self.i] + self.S[self.j]) % 256]])

        return keystream


import os


def _version():
    with open(os.path.join(os.path.dirname(__file__), '../VERSION')) as version_file:
        version = version_file.read().strip()

    return version


VERSION = _version()


from samson.utilities.manipulation import xor_buffs
from samson.utilities.bytes import Bytes
from samson.analyzers.analyzer import Analyzer
from copy import deepcopy
import struct

import logging
log = logging.getLogger(__name__)

class XORTranspositionAttack(object):
    """
    Performs a plaintext recovery attack.

    Using only an Analyzer, attempts to break a many-time pad using structural properties.
    This attack has two phases:
        * Transposition analysis - creates a matrix out of the bytes and transposes it so the rows share a common keystream byte before analyzing it. WARNING: the ciphertexts are trimmed to the length of the smallest sample.
        
        * Full-text analysis - reverts the transposed matrix and then copies the new, partial plaintexts over the original ciphertexts. Each ciphertext is analyzed in its entirety and the best character per position is chosen. This process can be repeated to incrementally recover more plaintext.
    
    Conditions:
        * A stream/OTP-like cipher is used. I.E. plaintext XOR keystream
        * The user has collected more than one ciphertext using the same keystream.
    """

    def __init__(self, analyzer: Analyzer):
        """
        Parameters:
            analyzer (Analyzer): Analyzer that correctly scores the underlying plaintext.
        """
        self.analyzer = analyzer


    def execute(self, ciphertexts: list, iterations: int=3) -> list:
        """
        Executes the attack.
        
        Parameters:
            ciphertexts (list): List of bytes-like ciphertexts using the same keystream.
            iterations   (int): Number of iterations of the full-text analysis phase. Accuracy-time trade-off.

        Returns:
            list: List of recovered plaintexts.
        """
        min_size = min([len(ciphertext) for ciphertext in ciphertexts])

        same_size_ciphers = [ciphertext[:min_size] for ciphertext in ciphertexts]
        transposed_ciphers = [bytearray(transposed) for transposed in zip(*same_size_ciphers)]
        assert [bytearray(transposed) for transposed in zip(*transposed_ciphers)] == same_size_ciphers

        log.debug("Starting initial transposition analysis")

        # Transposition analysis first (transposition)
        transposed_plaintexts = []
        for cipher in transposed_ciphers:
            all_chars = {}
            for char in range(256):
                plaintext = Bytes(struct.pack('B', char)).stretch(len(cipher)) ^ cipher

                all_chars[char] = (self.analyzer.analyze(plaintext), plaintext)

            transposed_plaintexts.append(sorted(all_chars.items(), key=lambda kv: kv[1][0], reverse=True)[0][1][1])


        retransposed_plaintexts = [bytearray(transposed) for transposed in zip(*transposed_plaintexts)]

        log.debug("Starting full-text analysis on retransposed text")

        # Clean up with a character-by-character, higher-context analysis (retransposed)
        for j in range(iterations):
            log.debug("Starting iteration {}/{}".format(j + 1, iterations))
            differential_mask = bytearray()

            for i in range(min_size):
                all_chars = {}

                for char in range(256):
                    full_text_analyses = []

                    frames = []
                    for curr_cipher in retransposed_plaintexts:
                        cipher_copy = deepcopy(curr_cipher)
                        cipher_copy[i] = ord(Bytes(struct.pack('B', char)) ^ struct.pack('B', curr_cipher[i]))

                        preprocessed_frame = self.analyzer.preprocess(cipher_copy)
                        frames.append(preprocessed_frame)
                        full_text_analyses.append(self.analyzer.analyze(cipher_copy))

                    all_chars[char] = (sum(full_text_analyses), char)

                best_char = sorted(all_chars.items(), key=lambda kv: kv[1][0], reverse=True)[0][1][1]
                differential_mask += struct.pack('B', best_char)
                
            retransposed_plaintexts = [xor_buffs(cipher, differential_mask) for cipher in retransposed_plaintexts]

        return retransposed_plaintexts


from samson.utilities.math import find_prime, mod_inv, gcd
from samson.utilities.bytes import Bytes


class Paillier(object):
    """
    Paillier public key cryptosystem

    Partially homomorphic encryption. Supports addition and multiplication.
    """

    def __init__(self, p: int=None, q: int=None):
        """
        Parameters:
            p (int): Large prime.
            q (int): Large prime.
        """
        self.p = p or find_prime(512)
        self.q = q or find_prime(512)
        self.n = self.p * self.q

        self.phi = (self.p - 1) * (self.q - 1)
        self.g = self.n + 1
        self.priv = mod_inv(self.phi, self.n)


    def L(self, x: int) -> int:
        return (x - 1) // self.n


    def __repr__(self):
        return f"<Paillier: priv={self.priv}, p={self.p}, q={self.q}, n={self.n}, g={self.g}, phi={self.phi}>"

    def __str__(self):
        return self.__repr__()



    def encrypt(self, plaintext: bytes) -> int:
        """
        Encrypts `plaintext`.

        Parameters:
            plaintext (bytes): Plaintext.
        
        Returns:
            int: Ciphertext.
        """
        m = Bytes.wrap(plaintext).to_int()
        assert m < self.n

        r = Bytes.random(self.n.bit_length() + 7 // 8).to_int()
        while gcd(r, self.n) != 1:
            r = Bytes.random(self.n.bit_length() + 7 // 8).to_int()

        n_sqr = self.n ** 2
        return pow(self.g, m, n_sqr) * pow(r, self.n, n_sqr)

    

    def decrypt(self, ciphertext: int) -> Bytes:
        """
        Decrypts `ciphertext` back into plaintext.

        Parameters:
            ciphertext (int): Ciphertext.
        
        Returns:
            Bytes: Decrypted plaintext.
        """
        return Bytes((self.L(pow(ciphertext, self.phi, self.n ** 2)) * self.priv) % self.n)



# https://github.com/XMPPwocky/nodebeefcl/blob/master/beef.py
# https://github.com/v8/v8/blob/ceade6cf239e0773213d53d55c36b19231c820b5/src/js/math.js#L143
# https://v8.dev/blog/math-random <-- Looks to be wrong
# http://www.helsbreth.org/random/rng_mwc1616.html
class MWC1616(object):
    """
    Multyply-with-carry 1616
    """

    def __init__(self, seed: (int, int), a: int=18030, b: int=30903, bits: int=32):
        """
        Parameters:
            seed ((int, int)): An integer or two-tuple of integers. If just an integer, it will be split into two.
            a           (int): Multiplier for the state's first item.
            b           (int): Multiplier for the state's second item.
            bits        (int): Number of bits to allow in the internal state. We'll create a mask out of this.
        """
        if type(seed) == int:
            seed = ((seed >> 16) & 0xFFFF, seed & 0xFFFF)

        self.state = seed
        self.a = a
        self.b = b
        self.bits = bits
        self._mask = 2**bits - 1
    

    def __repr__(self):
        return f"<MWC1616: state={self.state}, a={self.a}, b={self.b}, bits={self.bits}>"


    def __str__(self):
        return self.__repr__()


    def generate(self) -> int:
        """
        Generates the next psuedorandom output.

        Returns:
            int: Next psuedorandom output.
        """
        s0, s1 = self.state
        s0 = (self.a * (s0 & 0xFFFF) + (s0 >> 16)) & self._mask
        s1 = (self.b * (s1 & 0xFFFF) + (s1 >> 16)) & self._mask

        self.state = (s0, s1)

        return ((s0 << 16) + (s1 & 0xFFFF)) & self._mask


from sympy import Poly
from samson.utilities.encoding import poly_to_int
from samson.prngs.glfsr import GLFSR

class FLFSR(GLFSR):
    """
    Fibonacci linear-feedback shift register.
    """

    def __init__(self, seed: int, polynomial: Poly):
        """
        Parameters:
            seed        (int): Initial value.
            polynomial (Poly): Either a `sympy` `Poly` or an integer that represents the polynomal.
        """
        self.state = seed

        if type(polynomial) is Poly:
            polynomial = poly_to_int(polynomial)

        self.polynomial = polynomial
        self.mask = 1
        self.wrap_around_mask = 2 ** (polynomial.bit_length()) - 1
        self.state &= self.wrap_around_mask


        poly_mask = polynomial
        while poly_mask:
            if poly_mask & self.mask:
                poly_mask ^= self.mask
            
            if not poly_mask:
                break
            
            self.mask <<= 1
        
        self.mask -= 1


    def __repr__(self):
        return f"<FLFSR: state={self.state}, polynomial={self.polynomial}, mask={self.mask}>"
    

    def __str__(self):
        return self.__repr__()



    def clock(self) -> int:
        """
        Generates the next psuedorandom output.

        Returns:
            int: Next psuedorandom output.
        """
        self.state <<= 1

        lsb = sum([int(bit) for bit in bin(self.state & self.polynomial)[2:]]) % 2
        self.state |= lsb
        self.state &= 0xFFFFFFFFFFFFFFFF
        return self.state & 1
    


    def reverse_clock(self, output: int):
        """
        Clocks the state in reverse given the previous output.

        Parameters:
            output (int): Previous output.
        """
        raise NotImplementedError()



    def generate(self) -> int:
        """
        Calls self.clock(). Here for interface uniformity.

        Returns:
            int: Next psuedorandom output.
        """
        return self.clock()
    


from samson.utilities.bytes import Bytes
from samson.oracles.default_oracle import DefaultOracle

class CBCIVKeyEquivalenceAttack(object):
    """
    Performs a key-recovery attack.

    Some systems may opt to reduce the number of parameters to decrypt a ciphertext by
    setting the IV so it matches the key. Even assuming they don't include the IV with
    the ciphertext, we can recover the key under certain conditions.

    Conditions:
        * CBC is being used
        * The IV and the key are equivalent
        * The user has access to an oracle that returns the decrypted plaintext
    """
    def __init__(self, oracle: DefaultOracle, block_size: int=16):
        """
        Parameters:
            oracle  (DefaultOracle): An oracle that takes in a bytes-like object and returns its decryption.
            block_size        (int): The block size of the block cipher in bytes.
        """
        self.oracle = oracle
        self.block_size = block_size


    def execute(self, ciphertext: bytes) -> Bytes:
        """
        Executes the attack.
        
        Parameters:
            ciphertext (bytes): Bytes-like ciphertext generated by the system.

        Returns:
            Bytes: The key to the cipher.
        """
        edited_cipher = Bytes.wrap(bytes(ciphertext))

        edited_cipher[self.block_size:self.block_size * 2] = b'\x00' * self.block_size
        edited_cipher[self.block_size * 2:self.block_size * 3] = edited_cipher[:self.block_size]

        recovered_bytes = self.oracle.request(bytes(edited_cipher))
        blocks = Bytes(recovered_bytes).chunk(self.block_size)

        derived_iv = blocks[0] ^ blocks[2]
        return derived_iv


from samson.analyzers.analyzer import Analyzer
from samson.utilities.analysis import count_items, chisquare
from samson.auxiliary.tokenizer import Tokenizer
from samson.auxiliary.token_list_handler import TokenListHandler
import string
import os
import re

ascii_range = [10, 13] + list(range(20, 127))

with open(os.path.join(os.path.dirname(__file__), 'cracklib-small'), 'r') as f:
    wordlist = {word.strip(): 0 for word in f.readlines()}

first_letter_frequencies = {
    't': .15978,
    'a': .11682,
    'o': .07631,
    'i': .07294,
    's': .06686,
    'w': .05497,
    'c': .05238,
    'b': .04434,
    'p': .04319,
    'h': .04200,
    'f': .04027,
    'm': .03826,
    'd': .03174,
    'r': .02826,
    'e': .02799,
    'l': .02415,
    'n': .02284,
    'g': .01642,
    'u': .01183,
    'v': .00824,
    'y': .00763,
    'j': .00511,
    'k': .00456,
    'q': .00222,
    'x': .00045,
    'z': .00045
}

letter_freq = {
    'e': 0.1249,
    't': 0.0928,
    'a': 0.0804,
    'o': 0.0764,
    'i': 0.0757,
    'n': 0.0723,
    's': 0.0651,
    'r': 0.0628,
    'h': 0.0505,
    'l': 0.0407,
    'd': 0.0382,
    'c': 0.0334,
    'u': 0.0273,
    'm': 0.0251,
    'f': 0.0240,
    'p': 0.0214,
    'g': 0.0187,
    'w': 0.0168,
    'y': 0.0166,
    'b': 0.0148,
    'v': 0.0105,
    'k': 0.0054,
    'x': 0.0023,
    'j': 0.0016,
    'q': 0.0012,
    'z': 0.0009
}

most_common_words = {
    'the': .0714,
    'of': .0416,
    'and': .0304,
    'to': .0260,
    'in': .0227,
    'a': .0206,
    'is': .0113,
    'that': .0108,
    'for': .0088,
    'it': .0077,
    'as': .0077,
    'was': .0074,
    'with': .0070,
    'be': .0065,
    'by': .0063,
    'on': .0062,
    'not': .0061,
    'he': .0055,
    'i': .0052,
    'this': .0051,
    'are': .0050,
    'or': .0049,
    'his': .0049,
    'from': .0047,
    'at': .0046,
    'which': .0042,
    'but': .0038,
    'have': .0037,
    'an': .0037,
    'had': .0035,
    'they': .0033,
    'you': .0031,
    'were': .0031,
    'their': .0029,
    'one': .0029,
    'all': .0028,
    'we': .0028,
    'can': .0022,
    'her': .0022,
    'has': .0022,
    'there': .0022,
    'been': .0022,
    'if': .0021,
    'more': .0021,
    'when': .0020,
    'will': .0020,
    'would': .0020,
    'who': .0020,
    'so': .0019,
    'no': .0019
}

most_common_bigrams = {
    'TH': .0356,
    'HE': .0307,
    'IN': .0243,
    'ER': .0205,
    'AN': .0199,
    'RE': .0185,
    'ON': .0176,
    'AT': .0149,
    'EN': .0145,
    'ND': .0135,
    'TI': .0134,
    'ES': .0134,
    'OR': .0128,
    'TE': .0120,
    'OF': .0117,
    'ED': .0117,
    'IS': .0113,
    'IT': .0112,
    'AL': .0109,
    'AR': .0107,
    'ST': .0105,
    'TO': .0104,
    'NT': .0104,
    'NG': .0095,
    'SE': .0093,
    'HA': .0093,
    'AS': .0087,
    'OU': .0087,
    'IO': .0083,
    'LE': .0083,
    'VE': .0083,
    'CO': .0079,
    'ME': .0079,
    'DE': .0076,
    'HI': .0076,
    'RI': .0073,
    'RO': .0073,
    'IC': .0070,
    'NE': .0069,
    'EA': .0069,
    'RA': .0069,
    'CE': .0065,
    'LI': .0062,
    'CH': .0060,
    'LL': .0058,
    'BE': .0058,
    'MA': .0057,
    'SI': .0055,
    'OM': .0055,
    'UR': .0054
}



def _num_common_words(string):
    return sum([val * string.count(word) for word, val in most_common_words.items()]) / len(string)


def _num_common_letters(string):
    return sum([val * string.count(letter) for letter, val in letter_freq.items()]) / len(string)


def _num_common_first_letters(string):
    words = string.split(' ')
    return sum([first_letter_frequencies[word[0]] for word in words if len(word) > 0 and word[0] in first_letter_frequencies]) / len(words)


def _num_bigrams(string):
    return sum([val * string.count(bigram.lower()) for bigram, val in most_common_bigrams.items()]) / len(string)

TOKENIZER = Tokenizer([word for word, _ in wordlist.items() if len(word) > 2], TokenListHandler)

class EnglishAnalyzer(Analyzer):
    """
    Analyzer for English text.
    """
    def __init__(self):
        pass


    def analyze(self, in_bytes: bytes) -> float:
        """
        Takes in a bytes-like object and returns a relative score.

        Parameters:
            in_bytes (bytes): The bytes-like object to be "scored".
        
        Returns:
            float: The relative score of the object.
        """
        try:
            as_str = in_bytes.decode()
        except UnicodeDecodeError as _:
            return 0

        str_lower = as_str.lower()
        words = [word for word in re.split('[?.,! ]', str_lower) if word != '']
        word_freq = sum([1 for w in words if len(w) > 2 and len(w) < 8])


        alphabet_ratio = sum([1 for char in str_lower if char in string.ascii_lowercase]) / len(as_str)
        ascii_ratio = sum([1 for char in str_lower if ord(char) in ascii_range]) / len(as_str)

        common_words = _num_common_words(str_lower)
        first_letter_freq = _num_common_first_letters(str_lower)
        bigrams = _num_bigrams(str_lower)

        found_words = sum([len(word) ** 2 for word in TOKENIZER.tokenize([str_lower])])

        expected_english_distribution = {}
        for letter, freq in letter_freq.items():
            expected_english_distribution[bytes(letter, 'utf-8')] = freq * len(in_bytes)


        # We divide it by the `length*2` to normalize it since I empirically found that the chisquared of
        # a uniform distribution of `length` bytes tends towards it.
        chisquared_analysis = 1 / (chisquare(count_items(in_bytes), expected_english_distribution) / (len(in_bytes) * 2))

        return ((_num_common_letters(as_str.lower()) + 1) * (word_freq * 2 + 1)) * (((alphabet_ratio + 1) ** 5 - 1) * 60) * ((ascii_ratio + 1) ** 2 - 1) * (common_words + 1) * (first_letter_freq + 1) * (bigrams * 25 + 1) * (found_words + 1) * chisquared_analysis


    
    def preprocess(self, in_bytes: bytes, in_ciphers: bytes=None) -> dict:
        """
        Takes in a bytes-like object and returns the processed feature-space used in scoring.
        This is good for training statistical models.

        Parameters:
            in_bytes   (bytes): The bytes-like object to be "scored".
            in_ciphers (bytes): (Optional) If set, will check if `in_bytes` is a subset of `in_ciphers`.
                                Used for supervised machine learning.
        
        Returns:
            dictionary: Dictionary containing the processed features.
        """
        try:
            as_str = in_bytes.decode()
        except UnicodeDecodeError as _:
            return {
                'word_freq': 0.0,
                'alphabet_ratio': 0.0,
                'ascii_ratio': 0.0,
                'common_letters': 0.0,
                'common_words': 0.0,
                'first_letter_freq': 0.0,
                'bigrams': 0.0
            }


        words = as_str.split(' ')
        word_freq = sum([1 for w in words if len(w) > 2 and len(w) < 8]) / len(as_str)


        alphabet_ratio = sum([1 for char in as_str.lower() if char in string.ascii_lowercase]) / len(as_str)
        ascii_ratio = sum([1 for char in as_str.lower() if ord(char) in ascii_range]) / len(as_str)

        common_words = _num_common_words(as_str.lower())
        common_letters = _num_common_letters(as_str.lower())
        first_letter_freq = _num_common_first_letters(as_str.lower())
        bigrams = _num_bigrams(as_str.lower())

        return_dict = {
            'word_freq': word_freq,
            'alphabet_ratio': alphabet_ratio,
            'ascii_ratio': ascii_ratio,
            'common_letters': common_letters,
            'common_words': common_words,
            'first_letter_freq': first_letter_freq,
            'bigrams': bigrams
        }

        if in_ciphers != None:
            return_dict['is_correct'] = int(in_bytes in in_ciphers)

        return return_dict


from samson.utilities.encoding import int_to_bytes
from samson.oracles.padding_oracle import PaddingOracle
from samson.utilities.bytes import Bytes
from random import randint

import logging
log = logging.getLogger(__name__)

def _ceil(a, b):
    return (a + b - 1) // b


def _append_and_merge(new_a, new_b, intervals):
    for i, (a, b) in enumerate(intervals):
        if not (b < new_a or a > new_b):
            new_a = min(a, new_a)
            new_b = max(b, new_b)
            intervals[i] = new_a, new_b
            return
    
    intervals.append((new_a, new_b))


class PKCS1v15PaddingOracleAttack(object):
    """
    Performs a plaintext recovery attack.

    The PKCS#1 v1.5 padding oracle attack found by Daniel Bleichenbacher is an adaptive chosen-plaintext attack that
    takes advantage of an information leak through the validation of the plaintext's padding. Using RSA's homomorphic
    properties, the algorithm can iteratively converge on the correct plaintext.

    Conditions:
        * RSA is being used
        * PKCS#1 v1.5 padding is being used
        * The user has access to an oracle that allows abitrary plaintext input and leaks whether the padding is correct.
    """

    def __init__(self, oracle: PaddingOracle):
        """
        Parameters:
            oracle (PaddingOracle): An oracle that takes in an integer and returns whether the padding is correct.
        """
        self.oracle = oracle


    
    def execute(self, ciphertext: int, n: int, e: int, key_length: int) -> Bytes:
        """
        Executes the attack.

        Parameters:
            ciphertext (int): The ciphertext represented as an integer.
                     n (int): The RSA instance's modulus.
                     e (int): The RSA instance's public exponent.
            key_length (int): The the bit length of the RSA instance (2048, 4096, etc).
        
        Returns:
            Bytes: The ciphertext's corresponding plaintext.
        """
        key_byte_len = key_length // 8

        # Convenience variables
        B = 2 ** (8 * (key_byte_len - 2))

        # Initial values
        c = ciphertext
        c_0 = ciphertext
        M = [(2*B, 3*B - 1)]
        i = 1

        if not self.oracle.check_padding(c):
            log.debug("Initial padding not correct; attempting blinding")

            # Step 1: Blinding
            while True:
                s = randint(0, n - 1)
                c_0 = (c * pow(s, e, n)) % n

                if self.oracle.check_padding(c_0):
                    log.debug("Padding is now correct; blinding complete")
                    break

        # Step 2
        while True:
            log.debug("Starting iteration {}".format(i))
            log.debug("Current intervals: {}".format(M))
            # Step 2.a
            if i == 1:
                s = _ceil(n, 3*B)

                log.debug("Starting search at {}".format(s))

                while True:
                    c = c_0 * pow(s, e, n) % n
                    if self.oracle.check_padding(c):
                        break

                    s += 1
            # Step 2.b
            elif len(M) >= 2:
                log.debug("Intervals left: {}".format(M))
                while True:
                    s += 1
                    c = c_0 * pow(s, e, n) % n

                    if self.oracle.check_padding(c):
                        break
                        
            # Step 2.c
            elif len(M) == 1:
                log.debug("Only one interval")

                a, b = M[0]

                if a == b:
                    return Bytes(b'\x00' + int_to_bytes(a, 'big'))

                r = _ceil(2 * (b*s - 2*B), n)
                s = _ceil(2*B + r*n, b)

                while True:
                    c = c_0 * pow(s, e, n) % n
                    if self.oracle.check_padding(c):
                        break

                    s += 1
                    if s > (3*B + r*n) // a:
                        r += 1
                        s = _ceil(2*B + r*n, b)

            M_new = []

            for a, b in M:
                min_r = _ceil(a*s - 3*B + 1, n)
                max_r = (b*s - 2*B) // n

                for r in range(min_r, max_r + 1):
                    new_a = max(a, _ceil(2*B + r*n, s))
                    new_b = min(b, (3*B - 1 + r*n) // s)

                    if new_a > new_b:
                        raise Exception("Step 3: new_a > new_b? new_a: {} new_b: {}".format(new_a, new_b))


                    # Now we need to check for overlap between ranges and merge them
                    _append_and_merge(new_a, new_b, M_new)


            if len(M_new) == 0:
                raise Exception("There are zero intervals in 'M_new'")

            M = M_new
            i += 1


from samson.utilities.manipulation import right_rotate
from samson.utilities.bytes import Bytes
from samson.hashes.sha2 import H_512, H_256
from copy import deepcopy

SIGMA = [
    [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15],
    [14, 10, 4, 8, 9, 15, 13, 6, 1, 12, 0, 2, 11, 7, 5, 3],
    [11, 8, 12, 0, 5, 2, 15, 13, 10, 14, 3, 6, 7, 1, 9, 4],
    [7, 9, 3, 1, 13, 12, 11, 14, 2, 6, 5, 10, 4, 0, 15, 8],
    [9, 0, 5, 7, 2, 4, 10, 15, 14, 1, 11, 12, 6, 8, 3, 13],
    [2, 12, 6, 10, 0, 11, 8, 3, 4, 13, 7, 5, 15, 14, 1, 9],
    [12, 5, 1, 15, 14, 13, 4, 10, 0, 7, 6, 3, 9, 2, 8, 11],
    [13, 11, 7, 14, 12, 1, 3, 9, 5, 0, 15, 4, 8, 6, 2, 10],
    [6, 15, 14, 9, 11, 3, 0, 8, 12, 2, 13, 7, 1, 4, 10, 5],
    [10, 2, 8, 4, 7, 6, 1, 5, 15, 11, 9, 14, 3, 12, 13, 0]
]



class BLAKE2(object):
    """
    Cryptographic hash function based on ChaCha.
    """

    def __init__(self, key: bytes, desired_hash_len: int):
        """
        Parameters:
            key            (bytes): (Optional) Bytes-like object to key the hash.
            desired_hash_len (int): Desired output length.
        """
        self.key = key
        self.digest_size = desired_hash_len
        self.block_size = self.BLOCK_SIZE


    def padding_func(self, message: bytes) -> bytes:
        if len(message) % self.BLOCK_SIZE != 0 or len(message) == 0:
            message = message + b'\x00' * (self.BLOCK_SIZE - (len(message) % self.BLOCK_SIZE))
        return message


    # https://en.wikipedia.org/wiki/BLAKE_(hash_function)
    def mix(self, V_a, V_b, V_c, V_d, x, y):
        V_a = (V_a + V_b + x) & self.MASKBITS
        V_d = right_rotate(V_d ^ V_a, self.ROTATIONS[0], bits=self.WORD_SIZE)

        V_c = (V_c + V_d) & self.MASKBITS
        V_b = right_rotate(V_b ^ V_c, self.ROTATIONS[1], bits=self.WORD_SIZE)

        V_a = (V_a + V_b + y) & self.MASKBITS
        V_d = right_rotate(V_d ^ V_a, self.ROTATIONS[2], bits=self.WORD_SIZE)

        V_c = (V_c + V_d) & self.MASKBITS
        V_b = right_rotate(V_b ^ V_c, self.ROTATIONS[3], bits=self.WORD_SIZE)

        return V_a, V_b, V_c, V_d


    def compress(self, h, iv, chunk, t, is_last_block):
        V = [None] * 16
        V[:8] = h
        V[8:] = iv

        V[12] ^= t & self.MASKBITS
        V[13] ^= t >> self.WORD_SIZE

        if is_last_block:
            V[14] ^= self.MASKBITS

        m = [m_i.to_int() for m_i in chunk.chunk(self.WORD_SIZE // 8)]

        for i in range(self.ROUNDS):
            S = [None] * 16
            S = SIGMA[i % 10]

            V[0], V[4], V[8], V[12] = self.mix(V[0], V[4], V[8], V[12], m[S[0]], m[S[1]])
            V[1], V[5], V[9], V[13] = self.mix(V[1], V[5], V[9], V[13], m[S[2]], m[S[3]])
            V[2], V[6], V[10], V[14] = self.mix(V[2], V[6], V[10], V[14], m[S[4]], m[S[5]])
            V[3], V[7], V[11], V[15] = self.mix(V[3], V[7], V[11], V[15], m[S[6]], m[S[7]])


            V[0], V[5], V[10], V[15] = self.mix(V[0], V[5], V[10], V[15], m[S[8]], m[S[9]])
            V[1], V[6], V[11], V[12] = self.mix(V[1], V[6], V[11], V[12], m[S[10]], m[S[11]])
            V[2], V[7], V[8], V[13] = self.mix(V[2], V[7], V[8], V[13], m[S[12]], m[S[13]])
            V[3], V[4], V[9], V[14] = self.mix(V[3], V[4], V[9], V[14], m[S[14]], m[S[15]])
        
        h = [x ^ y for x, y in zip(h, V[:8])]
        h = [x ^ y for x, y in zip(h, V[8:])]

        return h




    def hash(self, message: bytes) -> Bytes:
        """
        Hashes the `message`.

        Parameters:
            message (bytes): Message to be hashed.
        
        Returns:
            Bytes: The hash digest.
        """
        message = Bytes(message, 'little')
        state = deepcopy(self.IV)

        last_block_size = len(message) % self.BLOCK_SIZE

        if last_block_size == 0 and len(message) > 0:
            last_block_size = self.BLOCK_SIZE

        state[0] ^= (0x0101 << 16) + (len(self.key) << 8) + (self.digest_size)

        if len(self.key) > 0:
            message = self.padding_func(self.key) + message

        padded_msg = self.padding_func(message)
        bytes_compressed = 0

        msg_chunks = padded_msg.chunk(self.BLOCK_SIZE)

        for i, chunk in enumerate(msg_chunks):
            is_last_block = i == (len(msg_chunks) - 1)
            bytes_compressed += last_block_size if is_last_block else self.BLOCK_SIZE
            state = self.compress(state, self.IV, chunk, bytes_compressed, is_last_block)
            

        return sum([Bytes(h, byteorder='little').zfill(self.WORD_SIZE // 8) for h in state])[:self.digest_size]





class BLAKE2b(BLAKE2):
    WORD_SIZE = 64
    MASKBITS = 0xFFFFFFFFFFFFFFFF
    ROUNDS = 12
    BLOCK_SIZE = 128
    IV = H_512
    ROTATIONS = [32, 24, 16, 63]

    def __init__(self, desired_hash_len=64, key=b''):
        """
        Parameters:
            key            (bytes): (Optional) Bytes-like object to key the hash.
            desired_hash_len (int): Desired output length.
        """
        super().__init__(key, desired_hash_len)


    def __repr__(self):
        return f"<BLAKE2b: iv={self.IV}, digest_size={self.digest_size}, key={self.key}>"


    def __str__(self):
        return self.__repr__()


class BLAKE2s(BLAKE2):
    WORD_SIZE = 32
    MASKBITS = 0xFFFFFFFF
    ROUNDS = 10
    BLOCK_SIZE = 64
    IV = H_256
    ROTATIONS = [16, 12, 8, 7]


    def __init__(self, desired_hash_len=32, key=b''):
        """
        Parameters:
            key            (bytes): (Optional) Bytes-like object to key the hash.
            desired_hash_len (int): Desired output length.
        """
        super().__init__(key, desired_hash_len)


    def __repr__(self):
        return f"<BLAKE2s: iv={self.IV}, digest_size={self.digest_size}, key={self.key}>"


    def __str__(self):
        return self.__repr__()